/*

 PE to Dex4u relocatable exe converter
 codedy by 0x4e71 (news4e71@yahoo.com)

*/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define DEX4U_RELOCATABLE_EXE_32 1
#define TRIM_TRIGGER "--$$EXE+CUT+HERE$$-"

IMAGE_DOS_HEADER doshdr ;
IMAGE_NT_HEADERS32 nthdr;
IMAGE_SECTION_HEADER sec;

typedef struct _RELOC_BYTE {
	struct _RELOC_BYTE *next;
	BYTE rel_offset;	
} RELOC_BYTE, *PRELOC_BYTE;

typedef struct _DEX_RELOC {
	struct _DEX_RELOC *next;
	DWORD offset;
	DWORD byte_cnt;
	PRELOC_BYTE bytes;
	PRELOC_BYTE last_byte;
} DEX_RELOC, *PDEX_RELOC;

PDEX_RELOC reloc_list_head = 0;
PDEX_RELOC reloc_list_tail = 0;
DWORD total_relocations_size = 0;

PDEX_RELOC new_dex_reloc(DWORD offset) {	
	PDEX_RELOC retval = (PDEX_RELOC) malloc(sizeof(DEX_RELOC));
	retval->next = 0;
	retval->offset = offset;
	retval->bytes = 0;
	retval->last_byte = 0;
	retval->byte_cnt = 0;
	total_relocations_size += 6;
	return retval;	
};

void link_dex_reloc(PDEX_RELOC dr) {	
	if (reloc_list_head == 0) {
		reloc_list_head = dr;
		reloc_list_tail = dr;
	} else {
		reloc_list_tail->next = dr;
		reloc_list_tail = dr;
	}
}

void add_reloc_byte(PDEX_RELOC dr, BYTE rel_offset) {
	PRELOC_BYTE nb = (PRELOC_BYTE) malloc(sizeof(RELOC_BYTE));
	nb->rel_offset = rel_offset;
	nb->next = 0;
	if (dr->bytes == 0) { 
		dr->bytes = nb;
		dr->last_byte = nb;
	} else {
		dr->last_byte->next = nb;
		dr->last_byte = nb;		
	}
	dr->byte_cnt++;
	total_relocations_size += 1;
}


void write_header(HFILE of, DWORD dex4u_raw_size, DWORD relocs_size, DWORD entry_offset) {	
	printf("Writing header\n");
	BYTE hdr[] = {'D','E','X', DEX4U_RELOCATABLE_EXE_32};
	_lwrite(of, (LPCSTR) hdr, sizeof(hdr));
	_lwrite(of, (LPCSTR) &dex4u_raw_size,  4);
	DWORD image_offset = sizeof(hdr) + 
			sizeof(dex4u_raw_size)+ 
			sizeof(image_offset) + 
			sizeof(entry_offset) + 
			relocs_size;
	_lwrite(of, (LPCSTR) &image_offset,  4);
	_lwrite(of, (LPCSTR) &entry_offset, 4);
}

void write_relocations(HFILE of) {
	printf("Writing relocations\n");
	if (reloc_list_head != 0) {
		PDEX_RELOC p = reloc_list_head;
		while (p != 0) {
			_lwrite(of, (LPCSTR) &(p->offset), 4);
			_lwrite(of, (LPCSTR) &(p->byte_cnt), 2); // endian dependent!!
			PRELOC_BYTE pb = p->bytes;
			while (pb != 0) {
				_lwrite(of, (LPCSTR) &(pb->rel_offset),1);
				pb = pb->next;
			}
			p = p->next;
		}		
	}
	DWORD marker = 0XFFFFFFFF;
	_lwrite(of,(LPCSTR) &marker, 4);
};


void write_image(HFILE of, BYTE *dex4u_data, DWORD dex4u_raw_size) {	
	printf("Writing binary image\n");
	_lwrite(of, (LPCSTR) dex4u_data, dex4u_raw_size);	
};

void show_help() {
	fprintf(stderr, "Usage: pe2dex input.exe output.drx [/TRIM]\n");
}

int main(int argc, char **argv) {
	int reloc_present = 0;
	int image_base = 0;
	int dex4u_raw_offset = 0;
	int dex4u_raw_size   = 0;
	int dex4u_RVA = 0;
	int reloc_raw_offset = 0;
	int reloc_raw_size   = 0;
	int reloc_limit      = 0;
	int do_trim = 0;
	char *infile_name = 0;
	char *outfile_name = 0;
	
	BYTE *dex4u_data = 0;
	BYTE *reloc_data = 0;
	
	if (argc<3) {
		show_help();
		return 1;
	}
	for (int i=1; i<argc; i++) {
		if (stricmp(argv[i],"/TRIM")==0) {
			do_trim = 1;
		} else if (infile_name == 0) {
			infile_name = argv[i];
		} else {
			outfile_name = argv[i];
		}		
	}
	if ((infile_name==0) || (outfile_name==0)){
		show_help();
		return 1;
	}
	HFILE infile = _lopen(argv[1],OF_READ);
	if (infile == HFILE_ERROR) {
		fprintf(stderr, "cannot open %s\n",argv[1]);
		return 1;
	}
	_lread(infile, (LPVOID) &doshdr, sizeof(doshdr));
	// printf("New header at offset %x\n",doshdr.e_lfanew);
  _llseek(infile, doshdr.e_lfanew, FILE_BEGIN);
  _lread(infile, (LPVOID) &nthdr, sizeof(nthdr));
  // printf("Magic: %s\n", &(nthdr.Signature));
	image_base = nthdr.OptionalHeader.ImageBase;
	// printf("image base: %x\n", image_base);  
	
  if (nthdr.OptionalHeader.NumberOfRvaAndSizes >= (IMAGE_DIRECTORY_ENTRY_BASERELOC+1)) {
  	if (nthdr.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size > 0 ) {
  		reloc_present = 1;
  		printf("Relocations found. Size of relocs %d\n",nthdr.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size);
  		reloc_limit = nthdr.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size;
  	} else {
  		printf("No relocations\n");
  	}
	}
	while(1) {
		_lread(infile, (LPVOID) &sec, sizeof(sec));
		if (sec.Name[0] == 0) {
			break;
		}
		printf("Section: %s  Raw data at: %x - Size %x - RVA: %x\n", sec.Name, sec.PointerToRawData, sec.SizeOfRawData, sec.VirtualAddress);
		if (strcmp((const char *)sec.Name, ".reloc")==0) {
			reloc_raw_offset = sec.PointerToRawData;
			reloc_raw_size = sec.SizeOfRawData;
		} else if (strcmp((const char*)sec.Name, ".dex4u")==0) {
			dex4u_raw_offset = sec.PointerToRawData;
			dex4u_raw_size = sec.SizeOfRawData;
			dex4u_RVA = sec.VirtualAddress;
		}
	}
	if (reloc_present && (reloc_raw_offset==0)) {
		fprintf(stderr, "cannot find .reloc section");
		_lclose(infile);
		return 1;		
	}
	if (dex4u_raw_offset==0) {
		fprintf(stderr, "cannot find .dex4u section");
		_lclose(infile);
		return 1;		
	}
	// compute entrypoint
	DWORD entry_offset = (nthdr.OptionalHeader.AddressOfEntryPoint) - dex4u_RVA;
	// read dex4u section
	dex4u_data = (BYTE *) malloc(dex4u_raw_size);
	_llseek(infile, dex4u_raw_offset, FILE_BEGIN);
	_lread(infile, (LPVOID) dex4u_data, dex4u_raw_size);

	// if needed, read reloc
	if (reloc_present) {	
		reloc_data = (BYTE *) malloc(reloc_raw_size);
		_llseek(infile, reloc_raw_offset, FILE_BEGIN);
		_lread(infile, (LPVOID) reloc_data, reloc_raw_size);
	}	
	_lclose(infile);
	// dump relocations, if needed
	if (reloc_present) {
		BYTE *relptr = reloc_data;
		BYTE *limit_address = reloc_data + reloc_limit;
		DWORD dex_offset = 0;
		DWORD prev_offset = 0;
		PDEX_RELOC dex_reloc = 0;
		while (relptr < limit_address) {
			DWORD r_RVA = *((DWORD *)relptr);
			relptr+=4;			
			DWORD r_size = *((DWORD *)relptr);
			relptr+=4;
			//printf("RELOC RVA %x - Size %x\n", r_RVA, r_size);
			//printf("buffer %x - limit address %x\n",relptr,limit_address);
			for (int i=0; i< (r_size-8)/2; i++) {
			 	WORD r_entry = *((WORD *)relptr);
			 	relptr+=2;
			 	if (r_entry != 0) {
			 		BYTE r_type = (r_entry >> 12) & 0x000F;
			 		DWORD r_offset = r_entry & 0x0FFF;
			 		DWORD target_offset = (r_RVA+r_offset)-dex4u_RVA;
			 		if (r_type != 3) {
			 			printf("Reloc entry type %x not supported!\n", r_type);
			 			return 1;
			 		}
			 		// get rid of of default address, make it zero based
			 		*((DWORD *)(dex4u_data+target_offset)) = ((*((DWORD *)(dex4u_data+target_offset)))-(image_base+dex4u_RVA));			 		
					int need_new_reloc = 0;
					if ((dex_reloc == 0) || ((target_offset - prev_offset) > 255)) {
						need_new_reloc = 1;
					}
					if (dex_reloc != 0) {
						// is the byte basket full?
						if (dex_reloc->byte_cnt == 0xFFFF) {
							need_new_reloc = 1;
						}
					}
					if (need_new_reloc){
						dex_reloc = new_dex_reloc(target_offset);
						add_reloc_byte(dex_reloc, 0);
						link_dex_reloc(dex_reloc);
					} else {
						add_reloc_byte(dex_reloc, target_offset - prev_offset);
					}
					prev_offset = target_offset;
				}
			}			
			// add space for end marker
			total_relocations_size += 4;
		}
	}	
	HFILE outfile = _lcreat(argv[2],0);
	if (outfile == HFILE_ERROR) {
		fprintf(stderr, "cannot open output file %s\n",argv[2]);
		return 1;
	}
	write_header(outfile, dex4u_raw_size, total_relocations_size, entry_offset);
	if (reloc_present) {
		write_relocations(outfile);
	}
	if (!do_trim) {
		write_image(outfile,dex4u_data, dex4u_raw_size);
	} else {
		printf("Trimming\n");
		char *s = (char*) dex4u_data;
		DWORD trimmed_size = dex4u_raw_size;
		for (int n=0; n<(dex4u_raw_size-(strlen(TRIM_TRIGGER)+1)); n++) {
			if (strncmp(s, TRIM_TRIGGER,strlen(TRIM_TRIGGER))==0) {
				trimmed_size = ((DWORD)s) - ((DWORD)dex4u_data);
			}
			s++;
		}
		write_image(outfile,dex4u_data, trimmed_size);
	}
	printf("Ready: %s\n", outfile_name);
	_lclose(outfile);
	return 0 ;
}

